"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const QueryBuilder_1 = require("../query-builder/QueryBuilder");
const PlainObjectToNewEntityTransformer_1 = require("../query-builder/transformer/PlainObjectToNewEntityTransformer");
const PlainObjectToDatabaseEntityTransformer_1 = require("../query-builder/transformer/PlainObjectToDatabaseEntityTransformer");
const FindOptionsUtils_1 = require("../find-options/FindOptionsUtils");
const QueryRunnerProvider_1 = require("../query-runner/QueryRunnerProvider");
const SubjectOperationExecutor_1 = require("../persistence/SubjectOperationExecutor");
const SubjectBuilder_1 = require("../persistence/SubjectBuilder");
/**
 * Repository is supposed to work with your entity objects. Find entities, insert, update, delete, etc.
 */
class Repository {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(connection, metadata, queryRunnerProvider) {
        this.connection = connection;
        this.metadata = metadata;
        this.queryRunnerProvider = queryRunnerProvider;
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Returns object that is managed by this repository.
     * If this repository manages entity from schema,
     * then it returns a name of that schema instead.
     */
    get target() {
        return this.metadata.target;
    }
    /**
     * Checks if entity has an id.
     * If entity contains compose ids, then it checks them all.
     */
    hasId(entity) {
        return this.metadata.hasId(entity);
    }
    /**
     * Creates a new query builder that can be used to build a sql query.
     */
    createQueryBuilder(alias, queryRunnerProvider) {
        return new QueryBuilder_1.QueryBuilder(this.connection, queryRunnerProvider)
            .select(alias)
            .from(this.metadata.target, alias);
    }
    /**
     * Creates a new entity instance or instances.
     * Can copy properties from the given object into new entities.
     */
    create(plainObjectOrObjects) {
        if (plainObjectOrObjects instanceof Array)
            return plainObjectOrObjects.map(object => this.create(object));
        const newEntity = this.metadata.create();
        if (plainObjectOrObjects) {
            const plainObjectToEntityTransformer = new PlainObjectToNewEntityTransformer_1.PlainObjectToNewEntityTransformer();
            plainObjectToEntityTransformer.transform(newEntity, plainObjectOrObjects, this.metadata);
        }
        return newEntity;
    }
    /**
     * Creates a new entity from the given plan javascript object. If entity already exist in the database, then
     * it loads it (and everything related to it), replaces all values with the new ones from the given object
     * and returns this new entity. This new entity is actually a loaded from the db entity with all properties
     * replaced from the new object.
     */
    preload(object) {
        const queryBuilder = this.createQueryBuilder(this.metadata.table.name);
        const plainObjectToDatabaseEntityTransformer = new PlainObjectToDatabaseEntityTransformer_1.PlainObjectToDatabaseEntityTransformer();
        return plainObjectToDatabaseEntityTransformer.transform(object, this.metadata, queryBuilder);
    }
    /**
     * Merges multiple entities (or entity-like objects) into a one new entity.
     */
    merge(...objects) {
        const newEntity = this.metadata.create();
        const plainObjectToEntityTransformer = new PlainObjectToNewEntityTransformer_1.PlainObjectToNewEntityTransformer();
        objects.forEach(object => plainObjectToEntityTransformer.transform(newEntity, object, this.metadata));
        return newEntity;
    }
    /**
     * Persists one or many given entities.
     */
    persist(entityOrEntities) {
        return __awaiter(this, void 0, void 0, function* () {
            // if multiple entities given then go throw all of them and save them
            if (entityOrEntities instanceof Array)
                return Promise.all(entityOrEntities.map(entity => this.persist(entity)));
            const queryRunnerProvider = this.queryRunnerProvider || new QueryRunnerProvider_1.QueryRunnerProvider(this.connection.driver, true);
            try {
                const databaseEntityLoader = new SubjectBuilder_1.SubjectBuilder(this.connection, queryRunnerProvider);
                yield databaseEntityLoader.persist(entityOrEntities, this.metadata);
                const executor = new SubjectOperationExecutor_1.SubjectOperationExecutor(this.connection, queryRunnerProvider);
                yield executor.execute(databaseEntityLoader.operateSubjects);
                return entityOrEntities;
            }
            finally {
                if (!this.queryRunnerProvider)
                    yield queryRunnerProvider.releaseReused();
            }
        });
    }
    /**
     * Removes one or many given entities.
     */
    remove(entityOrEntities) {
        return __awaiter(this, void 0, void 0, function* () {
            // if multiple entities given then go throw all of them and save them
            if (entityOrEntities instanceof Array)
                return Promise.all(entityOrEntities.map(entity => this.remove(entity)));
            const queryRunnerProvider = this.queryRunnerProvider || new QueryRunnerProvider_1.QueryRunnerProvider(this.connection.driver, true);
            try {
                const databaseEntityLoader = new SubjectBuilder_1.SubjectBuilder(this.connection, queryRunnerProvider);
                yield databaseEntityLoader.remove(entityOrEntities, this.metadata);
                const executor = new SubjectOperationExecutor_1.SubjectOperationExecutor(this.connection, queryRunnerProvider);
                yield executor.execute(databaseEntityLoader.operateSubjects);
                return entityOrEntities;
            }
            finally {
                if (!this.queryRunnerProvider)
                    yield queryRunnerProvider.releaseReused();
            }
        });
    }
    /**
     * Finds entities that match given conditions and/or find options.
     */
    find(conditionsOrFindOptions, options) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.createFindQueryBuilder(conditionsOrFindOptions, options)
                .getMany();
        });
    }
    /**
     * Finds entities that match given conditions.
     * Also counts all entities that match given conditions,
     * but ignores pagination settings (maxResults, firstResult) options.
     */
    findAndCount(conditionsOrFindOptions, options) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.createFindQueryBuilder(conditionsOrFindOptions, options)
                .getManyAndCount();
        });
    }
    /**
     * Finds first entity that matches given conditions and/or find options.
     */
    findOne(conditionsOrFindOptions, options) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.createFindQueryBuilder(conditionsOrFindOptions, options)
                .getOne();
        });
    }
    /**
     * Finds entities with ids.
     * Optionally find options can be applied.
     */
    findByIds(ids, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const qb = this.createFindQueryBuilder(undefined, options);
            return qb.andWhereInIds(ids).getMany();
        });
    }
    /**
     * Finds entity with given id.
     * Optionally find options can be applied.
     */
    findOneById(id, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const qb = this.createFindQueryBuilder(undefined, options);
            return qb.andWhereInIds([id]).getOne();
        });
    }
    /**
     * Executes a raw SQL query and returns a raw database results.
     */
    query(query) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryRunnerProvider = this.queryRunnerProvider || new QueryRunnerProvider_1.QueryRunnerProvider(this.connection.driver);
            const queryRunner = yield queryRunnerProvider.provide();
            try {
                return yield queryRunner.query(query); // await is needed here because we are using finally
            }
            finally {
                yield queryRunnerProvider.release(queryRunner);
            }
        });
    }
    /**
     * Wraps given function execution (and all operations made there) in a transaction.
     * All database operations must be executed using provided repository.
     */
    transaction(runInTransaction) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryRunnerProvider = this.queryRunnerProvider || new QueryRunnerProvider_1.QueryRunnerProvider(this.connection.driver, true);
            const queryRunner = yield queryRunnerProvider.provide();
            const transactionRepository = new Repository(this.connection, this.metadata, queryRunnerProvider);
            try {
                yield queryRunner.beginTransaction();
                const result = yield runInTransaction(transactionRepository);
                yield queryRunner.commitTransaction();
                return result;
            }
            catch (err) {
                yield queryRunner.rollbackTransaction();
                throw err;
            }
            finally {
                yield queryRunnerProvider.release(queryRunner);
                if (!this.queryRunnerProvider)
                    yield queryRunnerProvider.releaseReused();
            }
        });
    }
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    /**
     * Creates a query builder from the given conditions or find options.
     * Used to create a query builder for find* methods.
     */
    createFindQueryBuilder(conditionsOrFindOptions, options) {
        const findOptions = FindOptionsUtils_1.FindOptionsUtils.isFindOptions(conditionsOrFindOptions) ? conditionsOrFindOptions : options;
        const conditions = FindOptionsUtils_1.FindOptionsUtils.isFindOptions(conditionsOrFindOptions) ? undefined : conditionsOrFindOptions;
        const alias = findOptions ? findOptions.alias : this.metadata.table.name;
        const qb = this.createQueryBuilder(alias);
        // if find options are given then apply them to query builder
        if (findOptions)
            FindOptionsUtils_1.FindOptionsUtils.applyOptionsToQueryBuilder(qb, findOptions);
        // if conditions are given then apply them to query builder
        if (conditions) {
            Object.keys(conditions).forEach(key => {
                const name = key.indexOf(".") === -1 ? alias + "." + key : key;
                qb.andWhere(name + "=:" + key);
            });
            qb.addParameters(conditions);
        }
        return qb;
    }
}
exports.Repository = Repository;

//# sourceMappingURL=Repository.js.map
