"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const QueryRunnerProvider_1 = require("../query-runner/QueryRunnerProvider");
const Subject_1 = require("../persistence/Subject");
const RelationMetadata_1 = require("../metadata/RelationMetadata");
const QueryBuilder_1 = require("../query-builder/QueryBuilder");
/**
 * Repository for more specific operations.
 */
class SpecificRepository {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(connection, metadata, queryRunnerProvider) {
        this.connection = connection;
        this.metadata = metadata;
        if (queryRunnerProvider) {
            this.queryRunnerProvider = queryRunnerProvider;
        }
        else {
            this.queryRunnerProvider = new QueryRunnerProvider_1.QueryRunnerProvider(connection.driver);
        }
    }
    /**
     * Sets given relatedEntityId to the value of the relation of the entity with entityId id.
     * Should be used when you want quickly and efficiently set a relation (for many-to-one and one-to-many) to some entity.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    setRelation(relationName, entityId, relatedEntityId) {
        return __awaiter(this, void 0, void 0, function* () {
            const propertyName = this.metadata.computePropertyName(relationName);
            if (!this.metadata.hasRelationWithPropertyName(propertyName))
                throw new Error(`Relation ${propertyName} was not found in the ${this.metadata.name} entity.`);
            const relation = this.metadata.findRelationWithPropertyName(propertyName);
            // if (relation.isManyToMany || relation.isOneToMany || relation.isOneToOneNotOwner)
            //     throw new Error(`Only many-to-one and one-to-one with join column are supported for this operation. ${this.metadata.name}#${propertyName} relation type is ${relation.relationType}`);
            if (relation.isManyToMany)
                throw new Error(`Many-to-many relation is not supported for this operation. Use #addToRelation method for many-to-many relations.`);
            let table, values = {}, conditions = {};
            if (relation.isOwning) {
                table = relation.entityMetadata.table.name;
                values[relation.name] = relatedEntityId;
                conditions[relation.joinColumn.referencedColumn.name] = entityId;
            }
            else {
                table = relation.inverseEntityMetadata.table.name;
                values[relation.inverseRelation.name] = relatedEntityId;
                conditions[relation.inverseRelation.joinColumn.referencedColumn.name] = entityId;
            }
            const queryRunner = yield this.queryRunnerProvider.provide();
            yield queryRunner.update(table, values, conditions);
            yield this.queryRunnerProvider.release(queryRunner);
        });
    }
    /**
     * Sets given relatedEntityId to the value of the relation of the entity with entityId id.
     * Should be used when you want quickly and efficiently set a relation (for many-to-one and one-to-many) to some entity.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    setInverseRelation(relationName, relatedEntityId, entityId) {
        return __awaiter(this, void 0, void 0, function* () {
            const propertyName = this.metadata.computePropertyName(relationName);
            if (!this.metadata.hasRelationWithPropertyName(propertyName))
                throw new Error(`Relation ${propertyName} was not found in the ${this.metadata.name} entity.`);
            const relation = this.metadata.findRelationWithPropertyName(propertyName);
            // if (relation.isManyToMany || relation.isOneToMany || relation.isOneToOneNotOwner)
            //     throw new Error(`Only many-to-one and one-to-one with join column are supported for this operation. ${this.metadata.name}#${propertyName} relation type is ${relation.relationType}`);
            if (relation.isManyToMany)
                throw new Error(`Many-to-many relation is not supported for this operation. Use #addToRelation method for many-to-many relations.`);
            let table, values = {}, conditions = {};
            if (relation.isOwning) {
                table = relation.inverseEntityMetadata.table.name;
                values[relation.inverseRelation.name] = relatedEntityId;
                conditions[relation.inverseRelation.joinColumn.referencedColumn.name] = entityId;
            }
            else {
                table = relation.entityMetadata.table.name;
                values[relation.name] = relatedEntityId;
                conditions[relation.joinColumn.referencedColumn.name] = entityId;
            }
            const queryRunner = yield this.queryRunnerProvider.provide();
            yield queryRunner.update(table, values, conditions);
            yield this.queryRunnerProvider.release(queryRunner);
        });
    }
    /**
     * Adds a new relation between two entities into relation's many-to-many table.
     * Should be used when you want quickly and efficiently add a relation between two entities.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    addToRelation(relationName, entityId, relatedEntityIds) {
        return __awaiter(this, void 0, void 0, function* () {
            const propertyName = this.metadata.computePropertyName(relationName);
            if (!this.metadata.hasRelationWithPropertyName(propertyName))
                throw new Error(`Relation ${propertyName} was not found in the ${this.metadata.name} entity.`);
            const relation = this.metadata.findRelationWithPropertyName(propertyName);
            if (!relation.isManyToMany)
                throw new Error(`Only many-to-many relation supported for this operation. However ${this.metadata.name}#${propertyName} relation type is ${relation.relationType}`);
            const queryRunner = yield this.queryRunnerProvider.provide();
            const insertPromises = relatedEntityIds.map(relatedEntityId => {
                const values = {};
                if (relation.isOwning) {
                    values[relation.junctionEntityMetadata.columns[0].name] = entityId;
                    values[relation.junctionEntityMetadata.columns[1].name] = relatedEntityId;
                }
                else {
                    values[relation.junctionEntityMetadata.columns[1].name] = entityId;
                    values[relation.junctionEntityMetadata.columns[0].name] = relatedEntityId;
                }
                return queryRunner.insert(relation.junctionEntityMetadata.table.name, values);
            });
            yield Promise.all(insertPromises);
            yield this.queryRunnerProvider.release(queryRunner);
        });
    }
    /**
     * Adds a new relation between two entities into relation's many-to-many table from inverse side of the given relation.
     * Should be used when you want quickly and efficiently add a relation between two entities.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    addToInverseRelation(relationName, relatedEntityId, entityIds) {
        return __awaiter(this, void 0, void 0, function* () {
            const propertyName = this.metadata.computePropertyName(relationName);
            if (!this.metadata.hasRelationWithPropertyName(propertyName))
                throw new Error(`Relation ${propertyName} was not found in the ${this.metadata.name} entity.`);
            const relation = this.metadata.findRelationWithPropertyName(propertyName);
            if (!relation.isManyToMany)
                throw new Error(`Only many-to-many relation supported for this operation. However ${this.metadata.name}#${propertyName} relation type is ${relation.relationType}`);
            const queryRunner = yield this.queryRunnerProvider.provide();
            try {
                const insertPromises = entityIds.map(entityId => {
                    const values = {};
                    if (relation.isOwning) {
                        values[relation.junctionEntityMetadata.columns[0].name] = entityId;
                        values[relation.junctionEntityMetadata.columns[1].name] = relatedEntityId;
                    }
                    else {
                        values[relation.junctionEntityMetadata.columns[1].name] = entityId;
                        values[relation.junctionEntityMetadata.columns[0].name] = relatedEntityId;
                    }
                    return queryRunner.insert(relation.junctionEntityMetadata.table.name, values);
                });
                yield Promise.all(insertPromises);
            }
            finally {
                yield this.queryRunnerProvider.release(queryRunner);
            }
        });
    }
    /**
     * Removes a relation between two entities from relation's many-to-many table.
     * Should be used when you want quickly and efficiently remove a many-to-many relation between two entities.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    removeFromRelation(relationName, entityId, relatedEntityIds) {
        return __awaiter(this, void 0, void 0, function* () {
            const propertyName = this.metadata.computePropertyName(relationName);
            if (!this.metadata.hasRelationWithPropertyName(propertyName))
                throw new Error(`Relation ${propertyName} was not found in the ${this.metadata.name} entity.`);
            const relation = this.metadata.findRelationWithPropertyName(propertyName);
            if (!relation.isManyToMany)
                throw new Error(`Only many-to-many relation supported for this operation. However ${this.metadata.name}#${propertyName} relation type is ${relation.relationType}`);
            // check if given relation entity ids is empty - then nothing to do here (otherwise next code will remove all ids)
            if (!relatedEntityIds || !relatedEntityIds.length)
                return Promise.resolve();
            const qb = new QueryBuilder_1.QueryBuilder(this.connection, this.queryRunnerProvider)
                .delete()
                .fromTable(relation.junctionEntityMetadata.table.name, "junctionEntity");
            const firstColumnName = relation.isOwning ? relation.junctionEntityMetadata.columns[0].name : relation.junctionEntityMetadata.columns[1].name;
            const secondColumnName = relation.isOwning ? relation.junctionEntityMetadata.columns[1].name : relation.junctionEntityMetadata.columns[0].name;
            relatedEntityIds.forEach((relatedEntityId, index) => {
                qb.orWhere(`(${firstColumnName}=:entityId AND ${secondColumnName}=:relatedEntity_${index})`)
                    .setParameter("relatedEntity_" + index, relatedEntityId);
            });
            yield qb
                .setParameter("entityId", entityId)
                .execute();
        });
    }
    /**
     * Removes a relation between two entities from relation's many-to-many table.
     * Should be used when you want quickly and efficiently remove a many-to-many relation between two entities.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    removeFromInverseRelation(relationName, relatedEntityId, entityIds) {
        return __awaiter(this, void 0, void 0, function* () {
            const propertyName = this.metadata.computePropertyName(relationName);
            if (!this.metadata.hasRelationWithPropertyName(propertyName))
                throw new Error(`Relation ${propertyName} was not found in the ${this.metadata.name} entity.`);
            const relation = this.metadata.findRelationWithPropertyName(propertyName);
            if (!relation.isManyToMany)
                throw new Error(`Only many-to-many relation supported for this operation. However ${this.metadata.name}#${propertyName} relation type is ${relation.relationType}`);
            // check if given entity ids is empty - then nothing to do here (otherwise next code will remove all ids)
            if (!entityIds || !entityIds.length)
                return Promise.resolve();
            const qb = new QueryBuilder_1.QueryBuilder(this.connection, this.queryRunnerProvider)
                .delete()
                .from(relation.junctionEntityMetadata.table.name, "junctionEntity");
            const firstColumnName = relation.isOwning ? relation.junctionEntityMetadata.columns[1].name : relation.junctionEntityMetadata.columns[0].name;
            const secondColumnName = relation.isOwning ? relation.junctionEntityMetadata.columns[0].name : relation.junctionEntityMetadata.columns[1].name;
            entityIds.forEach((entityId, index) => {
                qb.orWhere(`(${firstColumnName}=:relatedEntityId AND ${secondColumnName}=:entity_${index})`)
                    .setParameter("entity_" + index, entityId);
            });
            yield qb.setParameter("relatedEntityId", relatedEntityId).execute();
        });
    }
    /**
     * Performs both #addToRelation and #removeFromRelation operations.
     * Should be used when you want quickly and efficiently and and remove a many-to-many relation between two entities.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    addAndRemoveFromRelation(relation, entityId, addRelatedEntityIds, removeRelatedEntityIds) {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all([
                this.addToRelation(relation, entityId, addRelatedEntityIds),
                this.removeFromRelation(relation, entityId, removeRelatedEntityIds)
            ]);
        });
    }
    /**
     * Performs both #addToRelation and #removeFromRelation operations.
     * Should be used when you want quickly and efficiently and and remove a many-to-many relation between two entities.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    addAndRemoveFromInverseRelation(relation, relatedEntityId, addEntityIds, removeEntityIds) {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all([
                this.addToInverseRelation(relation, relatedEntityId, addEntityIds),
                this.removeFromInverseRelation(relation, relatedEntityId, removeEntityIds)
            ]);
        });
    }
    /**
     * Removes entity with the given id.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    removeById(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const alias = this.metadata.table.name;
            const parameters = {};
            let condition = "";
            if (this.metadata.hasMultiplePrimaryKeys) {
                condition = this.metadata.primaryColumns.map(primaryColumn => {
                    parameters[primaryColumn.propertyName] = id[primaryColumn.propertyName];
                    return alias + "." + primaryColumn.propertyName + "=:" + primaryColumn.propertyName;
                }).join(" AND ");
            }
            else {
                condition = alias + "." + this.metadata.firstPrimaryColumn.propertyName + "=:id";
                parameters["id"] = id;
            }
            yield new QueryBuilder_1.QueryBuilder(this.connection, this.queryRunnerProvider)
                .delete()
                .from(this.metadata.target, alias)
                .where(condition, parameters)
                .execute();
        });
    }
    /**
     * Removes all entities with the given ids.
     * Note that event listeners and event subscribers won't work (and will not send any events) when using this operation.
     */
    removeByIds(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            const alias = this.metadata.table.name;
            const parameters = {};
            let condition = "";
            if (this.metadata.hasMultiplePrimaryKeys) {
                condition = ids.map((id, idIndex) => {
                    this.metadata.primaryColumns.map(primaryColumn => {
                        parameters[primaryColumn.propertyName + "_" + idIndex] = id[primaryColumn.propertyName];
                        return alias + "." + primaryColumn.propertyName + "=:" + primaryColumn.propertyName + "_" + idIndex;
                    }).join(" AND ");
                }).join(" OR ");
            }
            else {
                condition = alias + "." + this.metadata.firstPrimaryColumn.propertyName + " IN (:ids)";
                parameters["ids"] = ids;
            }
            yield new QueryBuilder_1.QueryBuilder(this.connection, this.queryRunnerProvider)
                .delete()
                .from(this.metadata.target, alias)
                .where(condition, parameters)
                .execute();
        });
    }
    /**
     * Finds all relation ids in the given entities.
     */
    findRelationIds(relationOrName, entityOrEntities, inIds, notInIds) {
        return __awaiter(this, void 0, void 0, function* () {
            const relation = this.convertMixedRelationToMetadata(relationOrName);
            if (!(entityOrEntities instanceof Array))
                entityOrEntities = [entityOrEntities];
            const entityReferencedColumn = relation.isOwning ? relation.joinTable.referencedColumn : relation.inverseRelation.joinTable.inverseReferencedColumn;
            const ownerEntityColumn = relation.isOwning ? relation.junctionEntityMetadata.columns[0] : relation.junctionEntityMetadata.columns[1];
            const inverseEntityColumn = relation.isOwning ? relation.junctionEntityMetadata.columns[1] : relation.junctionEntityMetadata.columns[0];
            let entityIds = this.convertEntityOrEntitiesToIdOrIds(entityReferencedColumn, entityOrEntities);
            if (!(entityIds instanceof Array))
                entityIds = [entityIds];
            // /*
            // filter out empty entity ids
            entityIds = entityIds.filter(entityId => entityId !== null && entityId !== undefined);
            // if no entity ids at the end, then we don't need to load anything
            if (entityIds.length === 0)
                return [];
            // */
            const ids = [];
            const promises = entityIds.map((entityId) => {
                const qb = new QueryBuilder_1.QueryBuilder(this.connection, this.queryRunnerProvider)
                    .select("junction." + inverseEntityColumn.name + " AS id")
                    .fromTable(relation.junctionEntityMetadata.table.name, "junction")
                    .andWhere("junction." + ownerEntityColumn.name + "=:entityId", { entityId: entityId });
                if (inIds && inIds.length > 0)
                    qb.andWhere("junction." + inverseEntityColumn.name + " IN (:inIds)", { inIds: inIds });
                if (notInIds && notInIds.length > 0)
                    qb.andWhere("junction." + inverseEntityColumn.name + " NOT IN (:notInIds)", { notInIds: notInIds });
                return qb.getScalarMany()
                    .then((results) => {
                    results.forEach(result => ids.push(result.id)); // todo: prepare result?
                });
            });
            yield Promise.all(promises);
            return ids;
        });
    }
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    /**
     * Converts entity or entities to id or ids map.
     */
    convertEntityOrEntitiesToIdOrIds(column, entityOrEntities) {
        if (entityOrEntities instanceof Array) {
            return entityOrEntities.map(entity => this.convertEntityOrEntitiesToIdOrIds(column, entity));
        }
        else {
            if (entityOrEntities instanceof Object) {
                return entityOrEntities[column.propertyName];
            }
            else {
                return entityOrEntities;
            }
        }
    }
    /**
     * Converts relation name, relation name in function into RelationMetadata.
     */
    convertMixedRelationToMetadata(relationOrName) {
        if (relationOrName instanceof RelationMetadata_1.RelationMetadata)
            return relationOrName;
        const relationName = relationOrName instanceof Function ? relationOrName(this.metadata.createPropertiesMap()) : relationOrName;
        return this.metadata.findRelationWithPropertyName(relationName);
    }
    /**
     * Extracts unique objects from given entity and all its downside relations.
     */
    extractObjectsById(entity, metadata, entityWithIds = []) {
        const promises = metadata.relations.map(relation => {
            const relMetadata = relation.inverseEntityMetadata;
            const value = relation.isLazy ? entity["__" + relation.propertyName + "__"] : entity[relation.propertyName];
            if (!value)
                return undefined;
            if (value instanceof Array) {
                const subPromises = value.map((subEntity) => {
                    return this.extractObjectsById(subEntity, relMetadata, entityWithIds);
                });
                return Promise.all(subPromises);
            }
            else {
                return this.extractObjectsById(value, relMetadata, entityWithIds);
            }
        });
        return Promise.all(promises.filter(result => !!result)).then(() => {
            if (!entityWithIds.find(entityWithId => entityWithId.entity === entity)) {
                const entityWithId = new Subject_1.Subject(metadata, entity);
                entityWithIds.push(entityWithId);
            }
            return entityWithIds;
        });
    }
}
exports.SpecificRepository = SpecificRepository;

//# sourceMappingURL=SpecificRepository.js.map
